/**
 * 2007, Digitalis Informatica. All rights reserved. Distribuicao e Gestao de
 * Informatica, Lda. Estrada de Paco de Arcos num.9 - Piso -1 2780-666 Paco de
 * Arcos Telefone: (351) 21 4408990 Fax: (351) 21 4408999
 * http://www.digitalis.pt
 */

package pt.digitalis.iss;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Properties;

import pt.digitalis.log.ILogWrapper;
import pt.digitalis.utils.common.collections.SearchableQueue;

/**
 * This is an implementation of the ISS interface.
 * 
 * Objects of this class perform the process scheduling management duties. <br>
 * It exposes methods to the client that allows process delegation, estimated
 * execution time inquiry and result fetching. <br>
 * Internally it manages the synchronously or asynchronously process execution
 * and a number of parameters that control the mode assignment to the processes.
 * <br>
 * The initial values of these parameters can be set externally in a properties
 * file. <br>
 * ISS is an acronym for <b>I</b>ntelligent <b>S</b>cheduling <b>S</b>ystem.
 * <br>
 * The asynchronicity threshold is reevaluated every time an asynchronous
 * request is served and every "N" received requests to do load balance. When
 * the ISS is serving requests asynchronously it is important to reevaluate
 * often the threshold to prevent further performance degrading.
 * 
 * @author Rodrigo Gonçalves <a
 *         href="mailto:rgoncalves@digitalis.pt">rgoncalves@digitalis.pt</a><br/>
 */
public class ISSImpl implements ISSRequestManagement, ISS {

	/** The requests that are waiting for asynchronous service. */
	private SearchableQueue<String, ServiceRequest> waitingProcesses = new SearchableQueue<String, ServiceRequest>();

	/**
	 * The requests that have been served asynchronously or that are in
	 * asynchronous service execution.
	 */
	private Map<String, ServiceRequest> servedProcesses = new HashMap<String, ServiceRequest>();

	/**
	 * This index stores the ids of the waiting, executing and served processes.
	 * It's purposes is to make sure that the same request is not served
	 * multiple times.
	 */
	private HashSet<String> processIndex = new HashSet<String>();

	/** The statistics engine used to calculate the ETE. */
	private ServiceStatistics waitingQueueStats = new ServiceStatistics();

	/**
	 * The statistics engine used to coordinate the asynchronous threshold
	 * updates
	 */
	private ServiceStatistics executionTimeStats = new ServiceStatistics();

	/** The logger wrapper. */
	private ILogWrapper logger;

	/**
	 * Counts the number of received requests since last reset. It is used to
	 * trigger the <code>asynchronocityThreshold</code> reevaluation.
	 */
	private int receivedRequestsCounter = 0;

	/** The number of requests being served synchronously. */
	private int numberOfRequestsInSynchronousService = 0;

	/**
	 * Threshold of asynchronous execution. If a new request arrives and the
	 * number of active workers is equal to this threshold, the incoming request
	 * will be put on the <code>waitingProcesses</code> and scheduled for
	 * asynchronous execution. Otherwise, the request is served immediately.
	 */
	private int asynchronicityThreshold;

	/**
	 * The minimum execution time (in milliseconds) needed to increase
	 * <code>asynchronicityThreshold</code>.
	 */
	private long minimumExecutionTime;

	/**
	 * The maximum execution time (in milliseconds)needed to decrease
	 * <code>asynchronicityThreshold</code>.
	 */
	private long maximumExecutionTime;

	/** The minimum number of requests to be served synchronously. */
	private int minimumThreshold;

	/** The maximum number of requests to be served synchronously. */
	private int maximumThreshold;

	/** The ISS identifier. */
	private String id = "";

	/** The number of maximum active workers. */
	private int maxActiveWorkers = 10;

	/** The number of active workers. */
	private int activeWorkers = 0;

	/** The entity responsible for the performance information logging. */
	private PerformanceLogger performanceLogger;

	/** Controls the logging of statistics data. */
	private boolean statisticsDumpEnabled;

	/**
	 * Every N received requests the <code>asynchronousThreshold</code> is
	 * reevaluated.
	 */
	final static int NUMBER_OF_RECEIVED_REQUESTS_TO_REEVALUATE_THRESHOLD = 5;

	/**
	 * Counts the number of asynchronously served requests since last reset. It
	 * is used to trigger the <code>asynchronocityThreshold</code>
	 * reevaluation.
	 */
	private int servedRequestsCounter = 0;

	/**
	 * Every N asynchronously served requests the
	 * <code>asynchronousThreshold</code> is reevaluated.
	 */
	final static int NUMBER_OF_SERVED_REQUESTS_TO_REEVALUATE_THRESHOLD = 10;

	/** A reevaluation counter for logging purposes */
	private int reevaluationCount = 0;

	/**
	 * Constructor. Builds a new ISS object with the given id and loads it's
	 * configuration from the supplied Properties object. If none is given then
	 * a default configuration is loaded.
	 * 
	 * @param id
	 *            the ISS id
	 * @param configurationParameters
	 *            the configuration parameters
	 * @param logger
	 *            the logger for ISS to use
	 * 
	 */
	public ISSImpl(String id, Properties configurationParameters,
			ILogWrapper logger) {
		this.id = id;
		this.statisticsDumpEnabled = true;
		this.loadConfiguration(configurationParameters);
		this.logger = logger;

		performanceLogger = new PerformanceLogger(this.logger);
	}

	/**
	 * Returns the ISS id.
	 * 
	 * @return the ISS id
	 */
	private String getID() {
		return this.id;
	}

	/**
	 * @see pt.digitalis.iss.ISS#delegate(IProcess, String)
	 * 
	 * Internally the method decides if the request is executed synchronously or
	 * asynchronously. If there's enough capacity to serve the request
	 * synchronously, it is executed immediately. Otherwise the request is
	 * queued for later execution. <br>
	 * <br>
	 * In synchronous execution mode the method returns the service request
	 * after the process execution has finished. In asynchronous mode it will
	 * place the newly created service request on the queue and return it with
	 * the queued state. <br>
	 * The method is also responsible for collecting statistical measures and
	 * submit them to the statistics engine. <br>
	 * <br>
	 * If an incoming request is already being served or has been served lately
	 * the request is returned with an updated status. <br>
	 * 
	 * <b>Implementation Note:</b> If this method is qualified with
	 * <code>synchronized</code> the client threads will not be able to call
	 * it if another client is executing it. In such situation, these threads
	 * will block waiting for the method's control and the ISS will operate in
	 * SYNCHRONOUS mode.
	 */
	public ServiceRequest delegate(IProcess aProcess, String id) {
		// Something to return...even if it's null...
		ServiceRequest request = null;

		// Check if the request has been served lately
		if (existsRequest(id))

			/*
			 * If the request is being served or has been served lately it is
			 * queued, and so it can be returned to the client. Prevents
			 * duplicated request delegation.
			 */
			request = fetchResults(id);
		else {
			boolean syncMode;

			incrementReceivedRequests();

			request = new ServiceRequest(aProcess, id);

			// Add the id to the index
			processIndex.add(request.getRequestID());

			// This block tests the capacity to serve synchronously. It's a sync
			// block so we can be sure that two requests don't get allowed to
			// execute in the same unique sync slot!
			synchronized (this) {
				syncMode = canExecuteMoreRequestsSynchronously();

				if (syncMode)
					// If in sync mode increment so that the next request sees
					// it immediately
					incrementNumberOfRequestsInSynchronousService();
			}

			// If there's available capacity serve immediately (SYNCHRONOUSLY)
			if (syncMode) {

				request.setState(ProcessState.EXECUTING);
				request.setBeginExecutionTime(System.currentTimeMillis());

				// Execute process (synchronously)
				ProcessResults results = aProcess.execute();
				request.setFinishExecutionTime(System.currentTimeMillis());

				// Add the total time in the system to the waitingQueueStats
				// engine
				waitingQueueStats.addSample(request.getFinishExecutionTime()
						- request.getArrivalTime());

				// Add the execution time to the executionStats
				executionTimeStats.addSample(request.getFinishExecutionTime()
						- request.getBeginExecutionTime());

				request.setResults(results.getResults());

				if (results.isOnError())
					request.setState(ProcessState.ERROR);
				else
					request.setState(ProcessState.FINISHED);

				// Checks if it's time to reevaluate the asynchronicityThreshold
				if (getReceivedRequestsCounter() >= NUMBER_OF_RECEIVED_REQUESTS_TO_REEVALUATE_THRESHOLD) {

					reevaluateAsynchronicityThreshold();
				}

				processIndex.remove(request.getRequestID());
				decrementNumberOfRequestsInSynchronousService();

				// Since the synchronous processes might have prevented
				// asynchronous workers to have been launched, check this out
				if (waitingProcesses.size() > 0)
					activateWorker();

			} else {

				// Otherwise, put request on the waiting queue (for asynchronous
				// service)
				request.setState(ProcessState.QUEUED);
				enqueue(request);
			}
		}

		return request;
	}

	/**
	 * @see pt.digitalis.iss.ISS#delegate(IProcess)
	 * 
	 * This form forces automatic id generation.
	 */
	public ServiceRequest delegate(IProcess process) {
		return this.delegate(process, null);
	}

	/**
	 * @see pt.digitalis.iss.ISS#existsRequest(String)
	 */
	public boolean existsRequest(String id) {
		return this.processIndex.contains(id);
	}

	/**
	 * This method is part of the public interface exposed to the Worker. It's
	 * purpose is to encapsulate the retrieval of a request waiting to be served
	 * from the <code>waitingProcesses</code>.
	 * 
	 * The method is synchronized to ensure queue consistency.
	 * 
	 * @see pt.digitalis.iss.ISSRequestManagement#getWaitingRequest()
	 */
	synchronized public ServiceRequest getWaitingRequest() {
		// Off the waiting queue
		ServiceRequest request = null;

		// If there are processes on the queue
		if (!waitingProcesses.isEmpty()) {

			// ...get the next...
			request = dequeue();
			request.setState(ProcessState.EXECUTING);

			// ...store on the served queue...
			store(request);
		}

		return request;
	}

	/**
	 * This method is part of the public interface exposed to the Worker. It's
	 * purpose is to serve as a callback method to the Worker. It's through this
	 * method that the asynchronously served request is returned to the ISS.
	 * <br>
	 * The returned request is stored on the ISS and statistical measures
	 * regarding the request service are collected and added to the
	 * waitingQueueStats engine. <br>
	 * If there are still requests waiting to be served the Worker is notified.
	 * <br>
	 * After a request is returned the <code>asynchronicityThreshold</code>
	 * must be reevaluated.
	 * 
	 * The method is synchronized to ensure queue consistency.
	 * 
	 * @see pt.digitalis.iss.ISSRequestManagement#returnRequest(pt.digitalis.iss.ServiceRequest)
	 */
	synchronized public void returnRequest(ServiceRequest returnedRequest) {
		// Put the returned request on the servedProcesses container
		store(returnedRequest);

		// Add the total time in the system to the waitingQueueStats engine
		waitingQueueStats.addSample(returnedRequest.getFinishExecutionTime()
				- returnedRequest.getArrivalTime());

		// Add the execution time to the executionStats
		executionTimeStats.addSample(returnedRequest.getFinishExecutionTime()
				- returnedRequest.getBeginExecutionTime());

		// If there are requests waiting to be served...
		if (!this.waitingProcesses.isEmpty()) {

			activateWorker();
		}

		this.increaseServedRequestsCounter();

		// Perform async threshold evaluation
		if (getServedRequestsCounter() >= NUMBER_OF_RECEIVED_REQUESTS_TO_REEVALUATE_THRESHOLD) {

			this.resetServedRequestCounter();
			reevaluateAsynchronicityThreshold();

		}
	}

	/**
	 * Returns the execution results to the Client.
	 * 
	 * @param requestID
	 *            the request id
	 * @return the request execution results
	 */
	public ServiceRequest fetchResults(String requestID) {
		// Fetch the results
		ServiceRequest result = fetchRequest(requestID);

		return result;
	}

	/**
	 * The method uses statistical information from the waitingQueueStats
	 * engine. If a client asks the ETE of a synchronously served request the
	 * method returns 0. This makes sense for two reasons:
	 * 
	 * 1) A synchronously served request client blocks while the request is
	 * being served.
	 * 
	 * 2) The request is not on the waiting queue so it will not have a "queue
	 * position"
	 * 
	 * This method uses waitingQueueStats as data source.
	 * 
	 * @param requestID
	 *            the id of the process
	 * 
	 * @return the estimated time of execution for the given process in
	 *         milliseconds
	 */
	public long getEstimatedTimeOfExecution(String requestID) {

		// Fetch results from waitingQueueStats
		double meanExecutionTime = waitingQueueStats.getMeanExecutionTime();
		long position = this.waitingProcesses.getPosition(requestID);

		// Init result
		long result = 0L;

		/*
		 * Evaluate result... If the request is NOT the next in line the ETE is
		 * equal to its position times the mean execution time
		 */
		if (position > 0L)
			result = (long) (position * meanExecutionTime);
		/*
		 * If the request is the next in line, the ETE is the mean execution
		 * time
		 */
		else
			result = (long) meanExecutionTime;

		return result;
	}

	/**
	 * This method manages the process request fetching operation. It searches
	 * for the process on both queues. If the request is waiting for execution
	 * or executing, it simply returns it to caller via the
	 * <code>fetchResults()</code> method. If the process has completed it's
	 * execution it is removed from the served queue and the index and then
	 * returned to the caller.
	 * 
	 * @param requestID
	 *            the request's ID
	 * @return the request associated with the given key
	 */
	private ServiceRequest fetchRequest(String requestID) {

		ServiceRequest result = null;

		// Search for request in the waiting queue
		if (this.waitingProcesses.containsKey(requestID))
			result = this.waitingProcesses.getValue(requestID);
		else {
			// If it's not on the waiting queue...
			// ...search on the served queue...
			result = this.servedProcesses.get(requestID);

			// ...if the process terminated it's execution...
			if (result != null
					&& (result.getState().equals(ProcessState.FINISHED) || result
							.getState().equals(ProcessState.ERROR))) {

				servedProcesses.remove(requestID);
				processIndex.remove(requestID);
			}
		}

		// Return...
		return result;
	}

	/**
	 * Puts the passed request on the waiting queue and notifies the worker of a
	 * new request to serve.
	 * 
	 * It's synchronized to ensure consistency on the waiting queue and on the
	 * Worker.
	 * 
	 * @param request
	 *            the request to put on the waiting queue
	 */
	synchronized private void enqueue(ServiceRequest request) {

		this.waitingProcesses.offer(request.getRequestID(), request);
		activateWorker();
	}

	/**
	 * Returns the next request to be served
	 * 
	 * @return the next request to be served
	 */
	private ServiceRequest dequeue() {
		return this.waitingProcesses.remove();
	}

	/**
	 * Stores the passed request into the served queue.
	 * 
	 * @param returned
	 *            the request to put on the served queue
	 */
	private void store(ServiceRequest returned) {
		this.servedProcesses.put(returned.getRequestID(), returned);
	}

	/**
	 * Returns the number of received requests.
	 * 
	 * @return the number of received requests since last reset
	 */
	private int getReceivedRequestsCounter() {
		return this.receivedRequestsCounter;
	}

	/**
	 * Increments the number of received requests by one unit.
	 */
	private void incrementReceivedRequests() {
		++this.receivedRequestsCounter;
	}

	/**
	 * Returns the maximum execution time for a request.
	 * 
	 * @return the maximumExecutionTime
	 */
	private long getMaximumExecutionTime() {
		return maximumExecutionTime;
	}

	/**
	 * Sets the maximum execution time for a request.
	 * 
	 * @param maximumExecutionTime
	 *            the maximumExecutionTime to set
	 */
	private void setMaximumExecutionTime(long maximumExecutionTime) {
		this.maximumExecutionTime = maximumExecutionTime;
	}

	/**
	 * Returns the maximum number of request in synchronous service
	 * 
	 * @return the maximumThreshold
	 */
	private int getMaximumThreshold() {
		return this.maximumThreshold;
	}

	/**
	 * Sets the maximum number of request in synchronous service
	 * 
	 * @param maximumThreshold
	 *            the maximumThreshold to set
	 */
	private void setMaximumThreshold(int maximumThreshold) {
		this.maximumThreshold = maximumThreshold;
	}

	/**
	 * Returns the minimum execution time a request can take.
	 * 
	 * @return the minimumExecutionTime
	 */
	private long getMinimumExecutionTime() {
		return this.minimumExecutionTime;
	}

	/**
	 * Sets the minimum execution time a request can take.
	 * 
	 * @param minimumExecutionTime
	 *            the minimumExecutionTime to set
	 */
	private void setMinimumExecutionTime(long minimumExecutionTime) {
		this.minimumExecutionTime = minimumExecutionTime;
	}

	/**
	 * Returns the minimum number of request in synchronous service
	 * 
	 * @return the minimumThreshold
	 */
	private int getMinimumThreshold() {
		return this.minimumThreshold;
	}

	/**
	 * Sets the minimum number of request in synchronous service .
	 * 
	 * @param minimumThreshold
	 *            the minimumThreshold to set
	 */
	private void setMinimumThreshold(int minimumThreshold) {
		this.minimumThreshold = minimumThreshold;
	}

	/**
	 * Returns the current number of request in synchronous service
	 * 
	 * @return the numberOfRequestsInSynchronousService
	 */
	private int getNumberOfRequestsInSynchronousService() {
		return this.numberOfRequestsInSynchronousService;
	}

	/** Increments <code>NumberOfRequestsInSynchronousService</code> by one. */
	private void incrementNumberOfRequestsInSynchronousService() {
		++this.numberOfRequestsInSynchronousService;
	}

	/** Decrements <code>NumberOfRequestsInSynchronousService</code> by one. */
	private void decrementNumberOfRequestsInSynchronousService() {
		--this.numberOfRequestsInSynchronousService;
	}

	/** Increments <code>asynchronicityThreshold</code> by one. */
	private void incrementAsynchronicityThreshold() {
		this.asynchronicityThreshold++;
	}

	/** Decrements <code>asynchronicityThreshold</code> by one. */
	private void decrementAsynchronicityThreshold() {
		this.asynchronicityThreshold--;
	}

	/**
	 * Returns the currently defined asynchronicity threshold.
	 * 
	 * @return the asynchronicityThreshold
	 */
	private int getAsynchronicityThreshold() {
		return this.asynchronicityThreshold;
	}

	/**
	 * Sets asynchronicity threshold.
	 * 
	 * @param asynchronicityThreshold
	 *            the asynchronicityThreshold to set
	 */
	private void setAsynchronicityThreshold(int asynchronicityThreshold) {
		this.asynchronicityThreshold = asynchronicityThreshold;
	}

	/**
	 * Reevaluates the asynchronicity threshold according to the defined QoS
	 * parameters.
	 * 
	 * The algorithm adds one to the current threshold if the new value is lower
	 * than the maximum threshold and the execution time is lower than the
	 * minimum execution time defined. Conversely, it subtracts one to the
	 * current threshold if the new value is higher than the minimum threshold
	 * and the execution time is higher than the maximum execution time.
	 * 
	 * The use of two execution times allows for a system with hysteresis.
	 * 
	 * This method uses executionTimeStats as data source.
	 */
	private void reevaluateAsynchronicityThreshold() {

		/*
		 * The execution time of a request or the mean execution time of a
		 * series of requests
		 */
		long executionTime = (long) executionTimeStats.getMeanExecutionTime();

		// Save previous async threshold value
		int previousAsyncThresholdValue = getAsynchronicityThreshold();

		// Evaluate threshold
		if (executionTime < getMinimumExecutionTime()
				&& getAsynchronicityThreshold() < getMaximumThreshold())

			incrementAsynchronicityThreshold();

		else if (executionTime > getMaximumExecutionTime()
				&& getAsynchronicityThreshold() > getMinimumThreshold())

			decrementAsynchronicityThreshold();

		if (isStatisticsDumpEnabled()) {
			reevaluationCount++;

			// Every 5 reevaluations we log the performance information
			if (reevaluationCount > 4) {
				reevaluationCount = 0;

				// Log performance information
				performanceLogger.logPerformanceInformation(
						previousAsyncThresholdValue,
						getAsynchronicityThreshold(), getMinimumThreshold(),
						getMaximumThreshold(),
						getNumberOfRequestsInSynchronousService(),
						getActiveWorkers(), this.waitingProcesses.size(),
						executionTime, getMinimumExecutionTime(),
						getMaximumExecutionTime());
			}
		}
	}

	/**
	 * Checks if the system has enough capacity to have another Worker running,
	 * and if so creates it.
	 * 
	 * Synchronized because <code>activeWorkers</code> is read and updated on
	 * the body.
	 */
	synchronized private void activateWorker() {

		// If there's still capacity to have more workers...
		if (canHaveMoreWorkers()) {

			increaseActiveWorkers();
			Worker aWorker = new Worker(this);
			aWorker.start();
		}
		// else, the existing workers will eventually serve all the requests
	}

	/**
	 * This method loads the the configuration properties from a file and
	 * initialize the attributes with the loaded values.
	 * 
	 * @param configurationParameters
	 *            the configuration parameters to load into the ISS
	 */
	private void loadConfiguration(Properties configurationParameters) {

		if (configurationParameters == null)
			configurationParameters = new Properties();

		// Read configuration values from Properties (or get the default value
		// if it's not defined)
		Integer numGroups = new Integer(configurationParameters.getProperty(
				this.getID() + ".stats.numberOfGroups", "5").toString());
		Integer elemGroup = new Integer(configurationParameters.getProperty(
				this.getID() + ".stats.elementsPerGroup", "20").toString());

		Integer asynchThresh = new Integer(configurationParameters.getProperty(
				this.getID() + ".iss.asynchronicityThreshold", "20").toString());
		Integer minThresh = new Integer(configurationParameters.getProperty(
				this.getID() + ".iss.minimumThreshold", "3").toString());
		Integer maxThresh = new Integer(configurationParameters.getProperty(
				this.getID() + ".iss.maximumThreshold", "50").toString());
		Long minExecTime = new Long(configurationParameters.getProperty(
				this.getID() + ".iss.minimumExecutionTime", "500").toString());
		Long maxExecTime = new Long(configurationParameters.getProperty(
				this.getID() + ".iss.maximumExecutionTime", "5000").toString());
		Integer maxActiveWorkers = new Integer(configurationParameters
				.getProperty(this.getID() + ".iss.maximumActiveWorkers", "50")
				.toString());

		// Set the configuration values on both the stats engines
		waitingQueueStats.setConfigNumberOfGroups(numGroups);
		waitingQueueStats.setConfigElementsPerGroup(elemGroup);

		executionTimeStats.setConfigNumberOfGroups(numGroups);
		executionTimeStats.setConfigElementsPerGroup(elemGroup);

		// Set the configuration values for this object
		this.setAsynchronicityThreshold(asynchThresh);
		this.setMinimumExecutionTime(minExecTime);
		this.setMaximumExecutionTime(maxExecTime);
		this.setMinimumThreshold(minThresh);
		this.setMaximumThreshold(maxThresh);
		this.setMaxActiveWorkers(maxActiveWorkers);
	}

	/**
	 * @see pt.digitalis.iss.ISS#isStatisticsDumpEnabled()
	 */
	public boolean isStatisticsDumpEnabled() {
		return statisticsDumpEnabled;
	}

	/**
	 * @see pt.digitalis.iss.ISS#setStatisticsDumpEnabled(boolean)
	 */
	public void setStatisticsDumpEnabled(boolean showStatistics) {
		this.statisticsDumpEnabled = showStatistics;
	}

	/**
	 * Synchronized to guarantee the <code>activeWorkers</code> consistency.
	 * 
	 * @see pt.digitalis.iss.ISSRequestManagement#notifyWorkerFinished()
	 */
	public void notifyWorkerFinished() {
		// Decrease the number of active workers
		decreaseActiveWorkers();
	}

	/**
	 * Returns the defined maximum number of parallel Workers. Takes into
	 * account the number of sync requests and the number of async workers
	 * 
	 * @return the maxActiveWorkers value
	 */
	private int getMaxActiveWorkers() {

		int maxLoadAvailable = (getAsynchronicityThreshold() - getNumberOfRequestsInSynchronousService());

		if (maxActiveWorkers > maxLoadAvailable)
			return maxLoadAvailable;
		else
			return maxActiveWorkers;
	}

	/**
	 * Sets a new value to the maximum number.
	 * 
	 * @param maxActiveWorkers
	 *            the new maxActiveWorkers value to set
	 */
	private void setMaxActiveWorkers(int maxActiveWorkers) {
		this.maxActiveWorkers = maxActiveWorkers;
	}

	/**
	 * Returns the number of active Workers.
	 * 
	 * @return the activeWorkers
	 */
	private synchronized int getActiveWorkers() {
		return activeWorkers;
	}

	/**
	 * Increase the number of active workers.
	 */
	private synchronized void increaseActiveWorkers() {
		this.activeWorkers++;
	}

	/**
	 * Decrease the number of active workers.
	 */
	private synchronized void decreaseActiveWorkers() {
		this.activeWorkers--;
	}

	/**
	 * Returns the value of the served requests counter.
	 * 
	 * @return the served requests counter value
	 */
	private int getServedRequestsCounter() {
		return servedRequestsCounter;
	}

	/** Increases the served request counter by one. */
	private void increaseServedRequestsCounter() {
		++servedRequestsCounter;
	}

	/** Resets the served request counter to zero. */
	private void resetServedRequestCounter() {
		servedRequestsCounter = 0;
	}

	/**
	 * Tests if we can launch more workers
	 * 
	 * @return true if the current number of workers is still bellow the current
	 *         limit
	 */
	public boolean canHaveMoreWorkers() {
		return (getActiveWorkers() < getMaxActiveWorkers());
	}

	/**
	 * Tests if we can serve more requests synchronously. Takes into account the
	 * running sync processes and the running async workers
	 * 
	 * @return true if the current number of workers + sync process is still
	 *         bellow the current limit
	 */
	public boolean canExecuteMoreRequestsSynchronously() {
		return (getNumberOfRequestsInSynchronousService() < (getAsynchronicityThreshold() - getActiveWorkers()));
	}

	/** @see pt.digitalis.iss.ISSRequestManagement#areTooManyWorkers() */
	public boolean areTooManyWorkers() {
		return (getActiveWorkers() > getMaxActiveWorkers());
	}

	/*
	 * Prints the configuration values on the System.out. Used for DEBUG.
	 * 
	 * @SuppressWarnings("unused") public void printConfiguration() {
	 * System.out.println("Configuration of the ISS: " + this.getID() +
	 * "\n[waitingQueueStats] Number of Groups: " +
	 * waitingQueueStats.getConfigNumberOfGroups() + "\n[waitingQueueStats]
	 * Elements per Group: " + waitingQueueStats.getConfigElementsPerGroup() +
	 * "\n[iss] Asynchronicity Threshold: " + this.getAsynchronicityThreshold() +
	 * "\n[iss] Minimum Threshold: " + this.getMinimumThreshold() + "\n[iss]
	 * Maximum Threshold: " + this.getMaximumThreshold() + "\n[iss] Minimum
	 * Execution Time: " + this.getMinimumExecutionTime() + "\n[iss] Maximum
	 * Execution Time: " + this.getMaximumExecutionTime() + "\n[iss]
	 * MaxActiveWorkers: " + this.getMaxActiveWorkers()); }
	 */
}
