/** 2007, Digitalis Informatica. All rights reserved.
 * 
 * Distribuicao e Gestao de Informatica, Lda.
 * Estrada de Paco de Arcos num.9 - Piso -1
 * 2780-666 Paco de Arcos
 * Telefone: (351) 21 4408990
 * Fax: (351) 21 4408999
 * http://www.digitalis.pt 
 */

package pt.digitalis.iss;

import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import pt.digitalis.log.ILogWrapper;
import pt.digitalis.log.LogWrapperLog4JImpl;

/**
 * This class serves as an interface between the clients and the different ISSs.
 * All accesses to the ISSs are made through the manager which is also
 * responsible for the ISSs creation if they don't exist yet.
 * 
 * @author Rodrigo Gonçalves <a
 *         href="mailto:rgoncalves@digitalis.pt">rgoncalves@digitalis.pt</a><br/>
 * 
 */
final public class ISSManager {

    /** The managed ISSs. */
    static private Map<String, ISS> managedISSs = new HashMap<String, ISS>();

    /** The logger instance */
    static private ILogWrapper logger;

    /**
     * Returns the ISS with the given id and configuration. If it doesn't exist
     * yet, it is created. <br>
     * The Properties file contains configuration parameters for the ISS. <br>
     * 
     * @param id
     *            the ISS id
     * @param configuration
     *            a Properties object with the desired configuration parameters
     *            for the ISS
     * @return the ISS with the matching id
     */
    synchronized static public ISS getISS(String id, Properties configuration) {

        // Fetch the appropriate ISS
        ISS theISS = managedISSs.get(id);

        // If it doesn't exist
        if (theISS == null) {
            // ...create it...
            theISS = createISS(id, configuration);
        }

        // return the ISS
        return theISS;
    }

    /**
     * Returns the ISS with the given id. If it doesn't exist yet, it is
     * created.
     * 
     * @param id
     *            the ISS id
     * @return the ISS with the given id
     */
    synchronized static public ISS getISS(String id) {
        return ISSManager.getISS(id, null);
    }

    /**
     * Creates a new ISS object and places it on the container.
     * 
     * @param id
     *            the id of the ISS that will be created
     * @param configuration
     *            a Properties object with the desired configuration parameters
     *            for the ISS
     * @return the newly created ISS
     */
    synchronized static private ISS createISS(String id,
            Properties configuration) {

        // Create new ISS
        ISS newISS = new ISSImpl(id, configuration, getLogger());

        // Store the new ISS
        managedISSs.put(id, newISS);

        return newISS;
    }

    /**
     * Noninstantiabilty must be enforced by providing a private constructor.
     */
    private ISSManager() {

    }

    /**
     * Returns the active logger.
     * 
     * @return the logger
     */
    public static ILogWrapper getLogger() {

        // If a logger has not been provided so far create a default logger.
        if (logger == null) {
            logger = new LogWrapperLog4JImpl();
        }

        return logger;
    }

    /**
     * Sets the active logger.
     * 
     * @param logger
     *            the logger to set
     */
    public static void setLogger(ILogWrapper logger) {
        ISSManager.logger = logger;
    }
}
