/**
 * 2007, Digitalis Informatica. All rights reserved. Distribuicao e Gestao de
 * Informatica, Lda. Estrada de Paco de Arcos num.9 - Piso -1 2780-666 Paco de
 * Arcos Telefone: (351) 21 4408990 Fax: (351) 21 4408999
 * http://www.digitalis.pt
 */

package pt.digitalis.iss;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.TimeZone;

import pt.digitalis.log.ILogWrapper;

/**
 * This entity logs the performance information of the ISS. <br>
 * It collects the performance data from the ISS and sends the data to a
 * LogWrapper. <br>
 * 
 * @author Rodrigo Gonçalves <a
 *         href="mailto:rgoncalves@digitalis.pt">rgoncalves@digitalis.pt</a><br/>
 */
final public class PerformanceLogger {

	/** The logger wrapper. */
	private ILogWrapper logger;

	/** The date format mask. */
	final static SimpleDateFormat DATE_FORMAT_MASK = new SimpleDateFormat(
			"EEEE, MMM dd yyyy '-' HH:mm:ss");

	/** A separator. */
	final static String LINE_SEPARATOR = "\n--------------------------------------------------------------------------------------------------------------\n";

	/**
	 * Constructor. Sets the the log level as info and configures the time zone.
	 * 
	 * @param logger
	 *            the logger instance
	 */
	public PerformanceLogger(ILogWrapper logger) {
		// Initialize logger
		this.logger = logger;

		// Pass the currently default time zone to the date format mask.
		DATE_FORMAT_MASK.setTimeZone(TimeZone.getDefault());
	}

	/**
	 * Logs the passed performance information. <br>
	 * Gets the current timestamp from the OS, composes a message with the
	 * passed performance information and sends the message to the logger.
	 * 
	 * @param previousAsynchronousThreshold
	 *            the previous asynchronicity threshold value
	 * @param currentAsynchronousThreshold
	 *            the current asynchronicity threshold value
	 * @param minAsynchronousThreshold
	 *            the minimum asynchronicity threshold value defined
	 * @param maxAsynchronousThreshold
	 *            the maximum asynchronicity threshold value defined
	 * @param numberOfSynchronousServiceRequests
	 *            the number of requests being served synchronously
	 * @param numberOfAsynchronousWorkers 
	 *            the number of workers handling asynchronous requests
	 * @param queueSize
	 *            the number of processes waiting on the queue
	 * @param meanExecutionTime
	 *            the current expected time for a request to be served
	 * @param minExecutionTime
	 *            the minimum defined service time
	 * @param maxExecutionTime
	 *            the maximum defined service time
	 */
	public void logPerformanceInformation(int previousAsynchronousThreshold,
			int currentAsynchronousThreshold, int minAsynchronousThreshold,
			int maxAsynchronousThreshold,
			int numberOfSynchronousServiceRequests,
			int numberOfAsynchronousWorkers, int queueSize,
			long meanExecutionTime, long minExecutionTime, long maxExecutionTime) {
		// Get the current time stamp
		String timeStamp = DATE_FORMAT_MASK.format(new Date());

		// Assemble the performance information on a string
		String message = "\n\nData collected at: "
				+ timeStamp
				+ LINE_SEPARATOR
				+ PerformanceLogger.getAsynchronousThresholdInfo(
						previousAsynchronousThreshold,
						currentAsynchronousThreshold, minAsynchronousThreshold,
						maxAsynchronousThreshold)
				+ "\n"
				+ PerformanceLogger
						.getSynchronousServicesInfo(numberOfSynchronousServiceRequests)
				+ "\n"
				+ PerformanceLogger
						.getAsynchronousWorkersInfo(numberOfAsynchronousWorkers)
				+ "\n"
				+ PerformanceLogger.getWaitingQueueInfo(queueSize)
				+ "\n"
				+ PerformanceLogger.getMeanServiceExecutionTimeInfo(
						meanExecutionTime, minExecutionTime, maxExecutionTime)
				+ LINE_SEPARATOR;

		logger.info(message);
	}

	/* --- HELPER METHODS --- */

	/**
	 * Composes a string with information on the asynchronicity threshold.
	 * 
	 * @param previousAsynchronousThreshold
	 *            the current asynchronicity threshold value
	 * @param currentAsynchronousThreshold
	 *            the current asynchronicity threshold value
	 * @param minAsynchronousThreshold
	 *            the minimum asynchronicity threshold value defined
	 * @param maxAsynchronousThreshold
	 *            the maximum asynchronicity threshold value defined
	 * 
	 * @return a string with the ISS asynchronous information
	 */
	static private String getAsynchronousThresholdInfo(
			int previousAsynchronousThreshold,
			int currentAsynchronousThreshold, int minAsynchronousThreshold,
			int maxAsynchronousThreshold) {
		return "Asynchronicity Threshold was " + previousAsynchronousThreshold
				+ " and now is " + currentAsynchronousThreshold + " ["
				+ minAsynchronousThreshold + "-" + maxAsynchronousThreshold
				+ "]";
	}

	/**
	 * Composes a string with information on the synchronous services.
	 * 
	 * @param numberOfSynchronousServiceRequests
	 *            the number of requests being served synchronously
	 * @return a string with the ISS synchronous information
	 */
	static private String getSynchronousServicesInfo(
			int numberOfSynchronousServiceRequests) {
		return "Number of requests in synchronous service: "
				+ numberOfSynchronousServiceRequests;
	}

	/**
	 * Composes a string with information on the asynchronous workers.
	 * 
	 * @param numberOfAsynchronousWorkers
	 *            the number of workers handling asynchronous requests
	 * @return a string with the ISS asynchronous information
	 */
	static private String getAsynchronousWorkersInfo(
			int numberOfAsynchronousWorkers) {
		return "Number of workers handling asynchronous requests: "
				+ numberOfAsynchronousWorkers;
	}

	/**
	 * Composes a string with information on the waiting queue status.
	 * 
	 * @param queueSize
	 *            the number of processes waiting on the queue
	 * 
	 * @return a string with the ISS waiting queue information
	 */
	static private String getWaitingQueueInfo(int queueSize) {
		return "Waiting queue size: " + queueSize;
	}

	/**
	 * Composes a string with information on the mean service execution time.
	 * 
	 * @param meanExecutionTime
	 *            the current expected time for a request to be served
	 * @param minExecutionTime
	 *            the minimum defined service time
	 * @param maxExecutionTime
	 *            the maximum defined service time
	 * @return a string with the ISS mean service execution time information
	 */
	static private String getMeanServiceExecutionTimeInfo(
			long meanExecutionTime, long minExecutionTime, long maxExecutionTime) {
		return "Mean service execution time: " + meanExecutionTime + " (ms) ["
				+ minExecutionTime + "-" + maxExecutionTime + "]";
	}
}
