/** 2007, Digitalis Informatica. All rights reserved.
 * 
 * Distribuicao e Gestao de Informatica, Lda.
 * Estrada de Paco de Arcos num.9 - Piso -1
 * 2780-666 Paco de Arcos
 * Telefone: (351) 21 4408990
 * Fax: (351) 21 4408999
 * http://www.digitalis.pt 
 */

package pt.digitalis.iss;

import java.util.HashMap;
import java.util.Map;

/**
 * This class holds information about a service request made to the ISS. It
 * stores the process to execute, it's state, start and end times and results.
 * 
 * @author Rodrigo Gonçalves <a
 *         href="mailto:rgoncalves@digitalis.pt">rgoncalves@digitalis.pt</a><br/>
 */
public class ServiceRequest {

    /** The process that will be executed. */
    private IProcess theProcess;

    /** The process state. */
    private ProcessState state;

    /** The time when the request arrived to the system. */
    private long arrivalTime;

    /** The time when the request began it's executing. */
    private long beginExecutionTime;

    /** The time when the request finished it's execution. */
    private long finishExecutionTime;

    /** Marks the request as asynchronous. */
    private boolean isAsynchronousRequest;

    /** The request id. */
    private String requestID;

    /** The process execution results. */
    private Map<String, Object> theResults = new HashMap<String, Object>();

    /**
     * Constructor. Builds a new Service Request for the passed IProcess. Also
     * sets the arrival time.
     * 
     * @param aProcess
     *            the process associated with this request
     * @param id
     *            the passed request id or a generated if it's null
     */
    public ServiceRequest(IProcess aProcess, String id) {

        this.arrivalTime = System.currentTimeMillis();
        this.theProcess = aProcess;

        // Manage id (if it's not supplied, generate a new one...)
        if (id == null)
            this.requestID = Integer.toString(this.hashCode());
        else
            this.requestID = id;
    }

    /**
     * Returns the time when the request finished it's execution.
     * 
     * @return the endTime
     */
    public long getFinishExecutionTime() {
        return finishExecutionTime;
    }

    /**
     * Sets the time when the request finished it's execution.
     * 
     * @param endTime
     *            the endTime to set
     */
    public void setFinishExecutionTime(long endTime) {
        this.finishExecutionTime = endTime;
    }

    /**
     * Returns the time when the request was made.
     * 
     * @return the startTime
     */
    public long getBeginExecutionTime() {
        return beginExecutionTime;
    }

    /**
     * Sets the time when the request was made.
     * 
     * @param beginExecutionTime
     *            the new value to set
     */
    public void setBeginExecutionTime(long beginExecutionTime) {
        this.beginExecutionTime = beginExecutionTime;
    }

    /**
     * Returns the request's state. <br>
     * To be used by the ISS client.
     * 
     * @return the state
     */
    public ProcessState getState() {
        return state;
    }

    /**
     * Set the request's state. If the state is set to QUEUED the request is
     * marked as asynchronous.
     * 
     * @param state
     *            the state to set
     */
    public void setState(ProcessState state) {
        // Set state
        this.state = state;

        // If process was QUEUED mark it as asynchronous
        if (this.state.equals(ProcessState.QUEUED))
            setAsynchronousRequest(true);
    }

    /**
     * Returns the process associated with this request.
     * 
     * @return the process
     */
    public IProcess getProcess() {
        return this.theProcess;
    }

    /**
     * Returns the request ID.
     * 
     * @return the requestID
     */
    public String getRequestID() {
        return requestID;
    }

    /**
     * Returns the Process execution results
     * 
     * @return the process execution results
     */
    public Map<String, Object> getResults() {
        return this.theResults;
    }

    /**
     * Sets the Process execution results
     * 
     * @param newResults
     *            the results to set
     */
    public void setResults(Map<String, Object> newResults) {
        this.theResults = newResults;
    }

    /**
     * Returns the Service Request representation as a String.
     * 
     * @return the request's String representation.
     */
    public String toString() {
        return "\n[ServiceRequest]\n  |--  Process: " + theProcess
                + "\n  |--  State: " + state + "\n  |--  Start Time: "
                + beginExecutionTime + "\n  |--  EndTime: "
                + finishExecutionTime + "\n  |--  RequestID: " + requestID
                + "\n  `--  Results: " + theResults + "\n";
    }

    /**
     * Arrival time inspector.
     * 
     * @return the arrivalTime value
     */
    public long getArrivalTime() {
        return arrivalTime;
    }

    /**
     * Arrival time modifier.
     * 
     * @param arrivalTime
     *            the arrivalTime value to set
     */
    public void setArrivalTime(long arrivalTime) {
        this.arrivalTime = arrivalTime;
    }

    /**
     * Asynchronous request flag inspector.
     * 
     * @return the isAsynchronousRequest value
     */
    public boolean isAsynchronousRequest() {
        return isAsynchronousRequest;
    }

    /**
     * Asynchronous request flag modifier. Access modifier is private to avoid
     * external setting.
     * 
     * @param isAsynchronousRequest
     *            the isAsynchronousRequest value to set
     */
    private void setAsynchronousRequest(boolean isAsynchronousRequest) {
        this.isAsynchronousRequest = isAsynchronousRequest;
    }
}
