/** 2007, Digitalis Informatica. All rights reserved.
 * 
 * Distribuicao e Gestao de Informatica, Lda.
 * Estrada de Paco de Arcos num.9 - Piso -1
 * 2780-666 Paco de Arcos
 * Telefone: (351) 21 4408990
 * Fax: (351) 21 4408999
 * http://www.digitalis.pt 
 */
package pt.digitalis.iss;

import java.util.LinkedList;
import java.util.Queue;

/**
 * This class is a small statistics engine that calculates the estimated time
 * for a request service.
 * 
 * The method used tries to reduce the number of insertion operations performed
 * on the container and iterations over it's elements. To achieve this, a method
 * that sums groups of samples is used. Every incoming sample is added to a
 * group total. When a certain number of samples is reached, the group total is
 * summed to the absolute total, the group total is stored on the container
 * (which also has it's oldest element removed) and the control variables reset.
 * 
 * @author Rodrigo Gonçalves <a
 *         href="mailto:rgoncalves@digitalis.pt">rgoncalves@digitalis.pt</a>
 */
final public class ServiceStatistics {

	/** Execution times data. */
	private Queue<Long> executionTimes = new LinkedList<Long>();

	/**
	 * The number of accumulated execution time groups to store on the
	 * container.
	 */
	private int configNumberOfGroups = 5;

	/** The number of elements in each group. */
	private int configElementsPerGroup = 20;

	/** The number of samples counter. */
	private int sampleCounter = 0;

	/** The number of processed groups. */
	private int numberOfProcessedGroups = 0;

	/** The absolute total time. */
	private Long total = new Long(0);

	/** The group total. */
	private Long groupTotal = new Long(0);

	/** Constructor */
	public ServiceStatistics() {
	}

	/**
	 * Adds a new sample to the statistics engine.
	 * 
	 * If the number of samples counter is equal to the number (max) of elements
	 * in a group the group total is added to the absolute total and stored in
	 * the container. The container and control variables are properly managed
	 * according to the system status.
	 * 
	 * Synchronized to ensure sample consistency (prevent management operations
	 * during concurrent access).
	 * 
	 * @param sample
	 *            the new sample to add
	 */
	synchronized public void addSample(long sample) {

		incrementSampleCounter();
		addToGroupTotal(sample);

		// If groupTotal accumulates the sum of defined number of samples in
		// group is time to do some house cleaning...
		if (getSampleCounter() == getConfigElementsPerGroup()) {
			// If there are less than configNumberOfGroups results on the
			// container
			// don't remove!!
			if (getExecutionTimes().size() == getConfigNumberOfGroups()) {
				setTotal(getTotal() - executionTimes.remove());
			}

			// If we haven't yet processed as many groups as defined on
			// <code>configNumberOfGroups</code>
			// we must update the number of processed groups
			if (getNumberOfProcessedGroups() < getConfigNumberOfGroups())
				setNumberOfProcessedGroups(getNumberOfProcessedGroups() + 1);

			// Enqueue the group total
			executionTimes.offer(getGroupTotal());
			// Add the group total to the absolute total
			setTotal(getTotal() + getGroupTotal());

			resetSampleCounter();
			resetGroupTotal();
		}
	}

	/**
	 * Calculates the mean service time of a request from the available data.
	 * 
	 * The evaluation considers two main situations: evaluating before and after
	 * the first group is processed. After the first group is processed, one
	 * must consider the case where the sample container is full or not.
	 * 
	 * Synchronized to ensure result consistency.
	 * 
	 * @return the mean service time of a request
	 */
	synchronized public double getMeanExecutionTime() {

		// Initialize result
		double result = 0.0;

		/*
		 * These variables are used to hold the terms as doubles. This is done
		 * because the evaluation includes a division of integers (which will
		 * produce a rational real number) and for code clarity purposes the
		 * casts are done outside of the formula.
		 */
		double groupTotal = (double) getGroupTotal();
		double sampleCounter = (double) getSampleCounter();
		double total = (double) getTotal();
		double numberOfGroups = (double) getConfigNumberOfGroups();
		double elementsPerGroup = (double) getConfigElementsPerGroup();
		double numberOfProcessedGroups = (double) getNumberOfProcessedGroups();

		// If we have completed the first group already...
		if (total != 0) {
			// If we have already processed the defined
			// <code>numberOfGroups</code>
			if (getNumberOfProcessedGroups() >= getConfigNumberOfGroups())
				result = (total + groupTotal)
						/ (numberOfGroups * elementsPerGroup + sampleCounter);
			// If we have lesser processed groups than what's defined on
			// <code>numberOfGroups</code>
			else
				result = (total + groupTotal)
						/ (numberOfProcessedGroups * elementsPerGroup + sampleCounter);

		}
		// ...if the first group is yet to be processed
		else {
			result = groupTotal / sampleCounter;
		}

		// Treat div-by-0 (otherwise it would return NaN)
		if (Double.isNaN(result))
			result = 0;

		return result;
	}

	/**
	 * Returns the execution times container.
	 * 
	 * @return the execution times container
	 */
	private Queue<Long> getExecutionTimes() {
		return this.executionTimes;
	}

	/**
	 * Returns the defined number of elements in a group.
	 * 
	 * @return the defined number of elements in a group
	 */
	public int getConfigElementsPerGroup() {
		return this.configElementsPerGroup;
	}

	/**
	 * Sets the number of elements in a group.
	 * 
	 * @param groupElementNumber
	 *            the configElementsPerGroup to set
	 */
	public void setConfigElementsPerGroup(int groupElementNumber) {
		this.configElementsPerGroup = groupElementNumber;
	}

	/**
	 * Returns the current group total value.
	 * 
	 * @return the groupTotal
	 */
	private Long getGroupTotal() {
		return groupTotal;
	}

	/**
	 * Adds an execution time to the group total.
	 * 
	 * @param executionTime
	 *            the execution time to add
	 */
	private void addToGroupTotal(Long executionTime) {
		this.groupTotal += executionTime;
	}

	/**
	 * Resets the group total.
	 */
	private void resetGroupTotal() {
		this.groupTotal = 0L; // NOTE: Obscure notation!!
	}

	/**
	 * Returns the current number of groups.
	 * 
	 * @return the configNumberOfGroups value
	 */
	public int getConfigNumberOfGroups() {
		return configNumberOfGroups;
	}

	/**
	 * Sets the number of groups.
	 * 
	 * @param numberOfGroups
	 *            the numberOfGroups to set
	 */
	public void setConfigNumberOfGroups(int numberOfGroups) {
		this.configNumberOfGroups = numberOfGroups;
	}

	/**
	 * Returns the number of samples elapsed.
	 * 
	 * @return the sampleCounter value
	 */
	private int getSampleCounter() {
		return sampleCounter;
	}

	/**
	 * Increments by one the number of elapsed samples.
	 */
	private void incrementSampleCounter() {
		this.sampleCounter++;
	}

	/**
	 * Resets the number of elapsed samples to 0. 
	 */
	private void resetSampleCounter() {
		this.sampleCounter = 0;
	}

	/**
	 * Returns the sample's absolute total.
	 * 
	 * @return the absolute total
	 */
	private Long getTotal() {
		return total;
	}

	/**
	 * Sets the absolute total value.
	 * 
	 * @param total
	 *            the new total to set
	 */
	private void setTotal(Long total) {
		this.total = total;
	}

	/**
	 * Returns the number of processed groups
	 * 
	 * @return the numberOfProcessedGroups
	 */
	private int getNumberOfProcessedGroups() {
		return numberOfProcessedGroups;
	}

	/**
	 * Sets the number of processed groups.
	 * 
	 * @param numberOfProcessedGroups
	 *            the numberOfProcessedGroups to set
	 */
	private void setNumberOfProcessedGroups(int numberOfProcessedGroups) {
		this.numberOfProcessedGroups = numberOfProcessedGroups;
	}

	/**
	 * This method enables that a client explicitly resets the stats engine. It
	 * clears all the sample records, and sets all the execution variables to
	 * zero.
	 */
	public void reset() {
		this.executionTimes.clear();
		this.resetGroupTotal();
		this.setNumberOfProcessedGroups(0);
		this.resetSampleCounter();
		this.setTotal(0L); // NOTE: Obscure notation!!
	}
}
