/** 2007, Digitalis Informatica. All rights reserved.
 * 
 * Distribuicao e Gestao de Informatica, Lda.
 * Estrada de Paco de Arcos num.9 - Piso -1
 * 2780-666 Paco de Arcos
 * Telefone: (351) 21 4408990
 * Fax: (351) 21 4408999
 * http://www.digitalis.pt 
 */

package pt.digitalis.iss;

/**
 * This class implements a Worker: a thread entity that executes requests.
 * 
 * When the worker is created and started it fetches a request from it's
 * associated ISS and serves and returns the request. The worker will repeat
 * this fetch-serve-return loop while there are requests on the ISS's waiting
 * queue. When the queue is empty the Worker notifies the ISS that it's duty is
 * over and ceases to exist.
 * 
 * The class is final so no one can change the way the Worker works (no pun
 * intended). If some changes were made to the Worker by subclassing the ISS
 * might not function properly. The Worker is a private part of the ISS and thus
 * not meant to be extended.
 * 
 * @author Rodrigo Gonçalves <a
 *         href="mailto:rgoncalves@digitalis.pt">rgoncalves@digitalis.pt</a><br/>
 */
final public class Worker extends Thread {

	/** The request to be executed by the worker. */
	private ServiceRequest theRequest;

	/** The ISS instance that launched this Worker (for callback). */
	private ISSRequestManagement theManager;

	/**
	 * Constructor. Constructs a new Worker and initialize ISS attribute.
	 * 
	 * @param theManager
	 *            the manager that called this Worker
	 */
	public Worker(ISSRequestManagement theManager) {
		// Initialize attribute
		this.theManager = theManager;
	}

	/**
	 * Overriding Thread#run()...
	 * 
	 * @see java.lang.Runnable#run()
	 */
	@Override
	public void run() {
		this.theRequest = this.theManager.getWaitingRequest();

		// Execute while there are requests to serve...
		while (this.theRequest != null) {

			this.theRequest.setBeginExecutionTime(System.currentTimeMillis());

			// ...execute the process, receiving the execution status
			ProcessResults result = this.theRequest.getProcess().execute();

			if (result.isOnError())
				this.theRequest.setState(ProcessState.ERROR);
			else
				this.theRequest.setState(ProcessState.FINISHED);

			this.theRequest.setFinishExecutionTime(System.currentTimeMillis());
			this.theRequest.setResults(result.getResults());

			// Return the served request to the manager
			this.theManager.returnRequest(this.theRequest);

			// Check is the current number of workers causes a violation of the
			// extended threshold every time a new process is finished
			if (theManager.areTooManyWorkers())
				// Too many workers, force the worker to die!
				this.theRequest = null;
			else
				// Tries to fetch a request in wait
				this.theRequest = this.theManager.getWaitingRequest();
		}

		this.theManager.notifyWorkerFinished();
	}
}
